# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.storage.filter import FilteringProxyStorage


def test_filtering_proxy_storage_content(sample_data):
    sample_content = sample_data['content'][0]
    storage = FilteringProxyStorage(storage={'cls': 'memory', 'args': {}})

    content = next(storage.content_get([sample_content['sha1']]))
    assert not content

    s = storage.content_add([sample_content])
    assert s == {
        'content:add': 1,
        'content:add:bytes': 48,
        'skipped_content:add': 0
    }

    content = next(storage.content_get([sample_content['sha1']]))
    assert content is not None

    s = storage.content_add([sample_content])
    assert s == {
        'content:add': 0,
        'content:add:bytes': 0,
        'skipped_content:add': 0
    }


def test_filtering_proxy_storage_revision(sample_data):
    sample_revision = sample_data['revision'][0]
    storage = FilteringProxyStorage(storage={'cls': 'memory', 'args': {}})

    revision = next(storage.revision_get([sample_revision['id']]))
    assert not revision

    s = storage.revision_add([sample_revision])
    assert s == {
        'revision:add': 1,
    }

    revision = next(storage.revision_get([sample_revision['id']]))
    assert revision is not None

    s = storage.revision_add([sample_revision])
    assert s == {
        'revision:add': 0,
    }


def test_filtering_proxy_storage_directory(sample_data):
    sample_directory = sample_data['directory'][0]
    storage = FilteringProxyStorage(storage={'cls': 'memory', 'args': {}})

    directory = next(storage.directory_missing([sample_directory['id']]))
    assert directory

    s = storage.directory_add([sample_directory])
    assert s == {
        'directory:add': 1,
    }

    directory = list(storage.directory_missing([sample_directory['id']]))
    assert not directory

    s = storage.directory_add([sample_directory])
    assert s == {
        'directory:add': 0,
    }



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import warnings

from swh.core.api import RPCClient

from ..exc import StorageAPIError


class RemoteStorage(RPCClient):
    """Proxy to a remote storage API"""
    api_exception = StorageAPIError

    def check_config(self, *, check_write):
        return self.post('check_config', {'check_write': check_write})

    def reset(self):
        return self.post('reset', {})

    def content_add(self, content):
        return self.post('content/add', {'content': content})

    def content_add_metadata(self, content):
        return self.post('content/add_metadata', {'content': content})

    def content_update(self, content, keys=[]):
        return self.post('content/update', {'content': content,
                                            'keys': keys})

    def content_missing(self, content, key_hash='sha1'):
        return self.post('content/missing', {'content': content,
                                             'key_hash': key_hash})

    def content_missing_per_sha1(self, contents):
        return self.post('content/missing/sha1', {'contents': contents})

    def skipped_content_missing(self, contents):
        return self.post('content/skipped/missing', {'contents': contents})

    def content_get(self, content):
        return self.post('content/data', {'content': content})

    def content_get_metadata(self, content):
        return self.post('content/metadata', {'content': content})

    def content_get_range(self, start, end, limit=1000):
        return self.post('content/range', {'start': start,
                                           'end': end,
                                           'limit': limit})

    def content_find(self, content):
        return self.post('content/present', {'content': content})

    def directory_add(self, directories):
        return self.post('directory/add', {'directories': directories})

    def directory_missing(self, directories):
        return self.post('directory/missing', {'directories': directories})

    def directory_ls(self, directory, recursive=False):
        return self.post('directory/ls', {'directory': directory,
                                          'recursive': recursive})

    def revision_get(self, revisions):
        return self.post('revision', {'revisions': revisions})

    def revision_log(self, revisions, limit=None):
        return self.post('revision/log', {'revisions': revisions,
                                          'limit': limit})

    def revision_shortlog(self, revisions, limit=None):
        return self.post('revision/shortlog', {'revisions': revisions,
                                               'limit': limit})

    def revision_add(self, revisions):
        return self.post('revision/add', {'revisions': revisions})

    def revision_missing(self, revisions):
        return self.post('revision/missing', {'revisions': revisions})

    def release_add(self, releases):
        return self.post('release/add', {'releases': releases})

    def release_get(self, releases):
        return self.post('release', {'releases': releases})

    def release_missing(self, releases):
        return self.post('release/missing', {'releases': releases})

    def object_find_by_sha1_git(self, ids):
        return self.post('object/find_by_sha1_git', {'ids': ids})

    def snapshot_add(self, snapshots):
        return self.post('snapshot/add', {'snapshots': snapshots})

    def snapshot_get(self, snapshot_id):
        return self.post('snapshot', {
            'snapshot_id': snapshot_id
        })

    def snapshot_get_by_origin_visit(self, origin, visit):
        return self.post('snapshot/by_origin_visit', {
            'origin': origin,
            'visit': visit
        })

    def snapshot_get_latest(self, origin, allowed_statuses=None):
        return self.post('snapshot/latest', {
            'origin': origin,
            'allowed_statuses': allowed_statuses
        })

    def snapshot_count_branches(self, snapshot_id):
        return self.post('snapshot/count_branches', {
            'snapshot_id': snapshot_id
        })

    def snapshot_get_branches(self, snapshot_id, branches_from=b'',
                              branches_count=1000, target_types=None):
        return self.post('snapshot/get_branches', {
            'snapshot_id': snapshot_id,
            'branches_from': branches_from,
            'branches_count': branches_count,
            'target_types': target_types
        })

    def origin_get(self, origins=None, *, origin=None):
        if origin is None:
            if origins is None:
                raise TypeError('origin_get expected 1 argument')
        else:
            assert origins is None
            origins = origin
            warnings.warn("argument 'origin' of origin_get was renamed "
                          "to 'origins' in v0.0.123.",
                          DeprecationWarning)
        return self.post('origin/get', {'origins': origins})

    def origin_search(self, url_pattern, offset=0, limit=50, regexp=False,
                      with_visit=False):
        return self.post('origin/search', {'url_pattern': url_pattern,
                                           'offset': offset,
                                           'limit': limit,
                                           'regexp': regexp,
                                           'with_visit': with_visit})

    def origin_count(self, url_pattern, regexp=False, with_visit=False):
        return self.post('origin/count', {'url_pattern': url_pattern,
                                          'regexp': regexp,
                                          'with_visit': with_visit})

    def origin_get_range(self, origin_from=1, origin_count=100):
        return self.post('origin/get_range', {'origin_from': origin_from,
                                              'origin_count': origin_count})

    def origin_add(self, origins):
        return self.post('origin/add_multi', {'origins': origins})

    def origin_add_one(self, origin):
        return self.post('origin/add', {'origin': origin})

    def origin_visit_add(self, origin, date, type):
        return self.post(
            'origin/visit/add',
            {'origin': origin, 'date': date, 'type': type})

    def origin_visit_update(self, origin, visit_id, status=None,
                            metadata=None, snapshot=None):
        return self.post('origin/visit/update', {'origin': origin,
                                                 'visit_id': visit_id,
                                                 'status': status,
                                                 'metadata': metadata,
                                                 'snapshot': snapshot})

    def origin_visit_upsert(self, visits):
        return self.post('origin/visit/upsert', {'visits': visits})

    def origin_visit_get(self, origin, last_visit=None, limit=None):
        return self.post('origin/visit/get', {
            'origin': origin, 'last_visit': last_visit, 'limit': limit})

    def origin_visit_find_by_date(self, origin, visit_date, limit=None):
        return self.post('origin/visit/find_by_date', {
            'origin': origin, 'visit_date': visit_date})

    def origin_visit_get_by(self, origin, visit):
        return self.post('origin/visit/getby', {'origin': origin,
                                                'visit': visit})

    def origin_visit_get_latest(self, origin, allowed_statuses=None,
                                require_snapshot=False):
        return self.post(
            'origin/visit/get_latest',
            {'origin': origin, 'allowed_statuses': allowed_statuses,
             'require_snapshot': require_snapshot})

    def stat_counters(self):
        return self.get('stat/counters')

    def refresh_stat_counters(self):
        return self.get('stat/refresh')

    def directory_entry_get_by_path(self, directory, paths):
        return self.post('directory/path', dict(directory=directory,
                                                paths=paths))

    def tool_add(self, tools):
        return self.post('tool/add', {'tools': tools})

    def tool_get(self, tool):
        return self.post('tool/data', {'tool': tool})

    def origin_metadata_add(self, origin_url, ts, provider, tool, metadata):
        return self.post('origin/metadata/add', {'origin_url': origin_url,
                                                 'ts': ts,
                                                 'provider': provider,
                                                 'tool': tool,
                                                 'metadata': metadata})

    def origin_metadata_get_by(self, origin_url, provider_type=None):
        return self.post('origin/metadata/get', {
            'origin_url': origin_url,
            'provider_type': provider_type
        })

    def metadata_provider_add(self, provider_name, provider_type, provider_url,
                              metadata):
        return self.post('provider/add', {'provider_name': provider_name,
                                          'provider_type': provider_type,
                                          'provider_url': provider_url,
                                          'metadata': metadata})

    def metadata_provider_get(self, provider_id):
        return self.post('provider/get', {'provider_id': provider_id})

    def metadata_provider_get_by(self, provider):
        return self.post('provider/getby', {'provider': provider})

    def diff_directories(self, from_dir, to_dir, track_renaming=False):
        return self.post('algos/diff_directories',
                         {'from_dir': from_dir,
                          'to_dir': to_dir,
                          'track_renaming': track_renaming})

    def diff_revisions(self, from_rev, to_rev, track_renaming=False):
        return self.post('algos/diff_revisions',
                         {'from_rev': from_rev,
                          'to_rev': to_rev,
                          'track_renaming': track_renaming})

    def diff_revision(self, revision, track_renaming=False):
        return self.post('algos/diff_revision',
                         {'revision': revision,
                          'track_renaming': track_renaming})



# Copyright (C) 2015-2016  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


class StorageDBError(Exception):
    """Specific storage db error (connection, erroneous queries, etc...)

    """

    def __str__(self):
        return 'An unexpected error occurred in the backend: %s' % self.args


class StorageAPIError(Exception):
    """Specific internal storage api (mainly connection)

    """

    def __str__(self):
        args = self.args
        return 'An unexpected error occurred in the api backend: %s' % args



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import glob
import pytest

from typing import Union

from pytest_postgresql import factories
from pytest_postgresql.janitor import DatabaseJanitor, psycopg2, Version

from os import path, environ
from hypothesis import settings
from typing import Dict

import swh.storage

from swh.core.utils import numfile_sortkey as sortkey

from swh.model.tests.generate_testdata import gen_contents, gen_origins


SQL_DIR = path.join(path.dirname(swh.storage.__file__), 'sql')

environ['LC_ALL'] = 'C.UTF-8'

DUMP_FILES = path.join(SQL_DIR, '*.sql')

# define tests profile. Full documentation is at:
# https://hypothesis.readthedocs.io/en/latest/settings.html#settings-profiles
settings.register_profile("fast", max_examples=5, deadline=5000)
settings.register_profile("slow", max_examples=20, deadline=5000)


@pytest.fixture
def swh_storage(postgresql_proc, swh_storage_postgresql):
    storage_config = {
        'cls': 'local',
        'args': {
            'db': 'postgresql://{user}@{host}:{port}/{dbname}'.format(
                host=postgresql_proc.host,
                port=postgresql_proc.port,
                user='postgres',
                dbname='tests'),
            'objstorage': {
                'cls': 'memory',
                'args': {}
            },
            'journal_writer': {
                'cls': 'memory',
            },
        },
    }
    storage = swh.storage.get_storage(**storage_config)
    return storage


@pytest.fixture
def swh_contents(swh_storage):
    contents = gen_contents(n=20)
    swh_storage.content_add(contents)
    return contents


@pytest.fixture
def swh_origins(swh_storage):
    origins = gen_origins(n=100)
    swh_storage.origin_add(origins)
    return origins


# the postgres_fact factory fixture below is mostly a copy of the code
# from pytest-postgresql. We need a custom version here to be able to
# specify our version of the DBJanitor we use.
def postgresql_fact(process_fixture_name, db_name=None, dump_files=DUMP_FILES):
    @pytest.fixture
    def postgresql_factory(request):
        """
        Fixture factory for PostgreSQL.

        :param FixtureRequest request: fixture request object
        :rtype: psycopg2.connection
        :returns: postgresql client
        """
        config = factories.get_config(request)
        if not psycopg2:
            raise ImportError(
                'No module named psycopg2. Please install it.'
            )
        proc_fixture = request.getfixturevalue(process_fixture_name)

        # _, config = try_import('psycopg2', request)
        pg_host = proc_fixture.host
        pg_port = proc_fixture.port
        pg_user = proc_fixture.user
        pg_options = proc_fixture.options
        pg_db = db_name or config['dbname']
        with SwhDatabaseJanitor(
                pg_user, pg_host, pg_port, pg_db, proc_fixture.version,
                dump_files=dump_files
        ):
            connection = psycopg2.connect(
                dbname=pg_db,
                user=pg_user,
                host=pg_host,
                port=pg_port,
                options=pg_options
            )
            yield connection
            connection.close()

    return postgresql_factory


swh_storage_postgresql = postgresql_fact('postgresql_proc')


# This version of the DatabaseJanitor implement a different setup/teardown
# behavior than than the stock one: instead of droping, creating and
# initializing the database for each test, it create and initialize the db only
# once, then it truncate the tables. This is needed to have acceptable test
# performances.
class SwhDatabaseJanitor(DatabaseJanitor):
    def __init__(
            self,
            user: str,
            host: str,
            port: str,
            db_name: str,
            version: Union[str, float, Version],
            dump_files: str = DUMP_FILES
    ) -> None:
        super().__init__(user, host, port, db_name, version)
        self.dump_files = sorted(
            glob.glob(dump_files), key=sortkey)

    def db_setup(self):
        with psycopg2.connect(
            dbname=self.db_name,
            user=self.user,
            host=self.host,
            port=self.port,
        ) as cnx:
            with cnx.cursor() as cur:
                for fname in self.dump_files:
                    with open(fname) as fobj:
                        sql = fobj.read().replace('concurrently', '').strip()
                        if sql:
                            cur.execute(sql)
            cnx.commit()

    def db_reset(self):
        with psycopg2.connect(
            dbname=self.db_name,
            user=self.user,
            host=self.host,
            port=self.port,
        ) as cnx:
            with cnx.cursor() as cur:
                cur.execute(
                    "SELECT table_name FROM information_schema.tables "
                    "WHERE table_schema = %s", ('public',))
                tables = set(table for (table,) in cur.fetchall())
                for table in tables:
                    cur.execute('truncate table %s cascade' % table)

                cur.execute(
                    "SELECT sequence_name FROM information_schema.sequences "
                    "WHERE sequence_schema = %s", ('public',))
                seqs = set(seq for (seq,) in cur.fetchall())
                for seq in seqs:
                    cur.execute('ALTER SEQUENCE %s RESTART;' % seq)
            cnx.commit()

    def init(self):
        with self.cursor() as cur:
            cur.execute(
                "SELECT COUNT(1) FROM pg_database WHERE datname=%s;",
                (self.db_name,))
            db_exists = cur.fetchone()[0] == 1
            if db_exists:
                cur.execute(
                    'UPDATE pg_database SET datallowconn=true '
                    'WHERE datname = %s;',
                    (self.db_name,))

        if db_exists:
            self.db_reset()
        else:
            with self.cursor() as cur:
                cur.execute('CREATE DATABASE "{}";'.format(self.db_name))
            self.db_setup()

    def drop(self):
        pid_column = 'pid'
        with self.cursor() as cur:
            cur.execute(
                'UPDATE pg_database SET datallowconn=false '
                'WHERE datname = %s;', (self.db_name,))
            cur.execute(
                'SELECT pg_terminate_backend(pg_stat_activity.{})'
                'FROM pg_stat_activity '
                'WHERE pg_stat_activity.datname = %s;'.format(pid_column),
                (self.db_name,))


@pytest.fixture
def sample_data() -> Dict:
    """Pre-defined sample storage object data to manipulate

    Returns:
        Dict of data (keys: content, directory, revision, person)

    """
    sample_content = {
        'blake2s256': b'\xbf?\x05\xed\xc1U\xd2\xc5\x168Xm\x93\xde}f(HO@\xd0\xacn\x04\x1e\x9a\xb9\xfa\xbf\xcc\x08\xc7',  # noqa
        'sha1': b'g\x15y+\xcb][\\\n\xf28\xb2\x0c_P[\xc8\x89Hk',
        'sha1_git': b'\xf2\xae\xfa\xba\xfa\xa6B\x9b^\xf9Z\xf5\x14\x0cna\xb0\xef\x8b',  # noqa
        'sha256': b"\x87\x022\xedZN\x84\xe8za\xf8'(oA\xc9k\xb1\x80c\x80\xe7J\x06\xea\xd2\xd5\xbeB\x19\xb8\xce",  # noqa
        'length': 48,
        'data': b'temp file for testing content storage conversion',
        'status': 'visible',
    }

    sample_content2 = {
        'blake2s256': b'\xbf?\x05\xed\xc1U\xd2\xc5\x168Xm\x93\xde}f(HO@\xd0\xacn\x04\x1e\x9a\xb9\xfa\xbf\xcc\x08\xc7',  # noqa
        'sha1': b'f\x15y+\xcb][\\\n\xf28\xb2\x0c_P[\xc8\x89Hk',
        'sha1_git': b'\xc2\xae\xfa\xba\xfa\xa6B\x9b^\xf9Z\xf5\x14\x0cna\xb0\xef\x8b',  # noqa
        'sha256': b"\x77\x022\xedZN\x84\xe8za\xf8'(oA\xc9k\xb1\x80c\x80\xe7J\x06\xea\xd2\xd5\xbeB\x19\xb8\xce",  # noqa
        'length': 50,
        'data': b'temp file for testing content storage conversion 2',
        'status': 'visible',
    }

    sample_directory = {
        'id': b'f\x15y+\xcb][\\\n\xf28\xb2\x0c_P[\xc8\x89Hk',
        'entries': []
    }

    sample_person = {
        'name': b'John Doe',
        'email': b'john.doe@institute.org',
        'fullname': b'John Doe <john.doe@institute.org>'
    }

    sample_revision = {
        'id': b'f\x15y+\xcb][\\\n\xf28\xb2\x0c_P[\xc8\x89Hk',
        'message': b'something',
        'author': sample_person,
        'committer': sample_person,
        'date': 1567591673,
        'committer_date': 1567591673,
        'type': 'tar',
        'directory': b'\xc2\xae\xfa\xba\xfa\xa6B\x9b^\xf9Z\xf5\x14\x0cna\xb0\xef\x8b',  # noqa
        'synthetic': False,
        'metadata': {},
        'parents': [],
    }

    return {
        'content': [sample_content, sample_content2],
        'person': [sample_person],
        'directory': [sample_directory],
        'revision': [sample_revision],
    }



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.storage import get_storage
from swh.storage.tests.test_storage import (  # noqa
    TestStorage, TestStorageGeneratedData)


# tests are executed using imported classes (TestStorage and
# TestStorageGeneratedData) using overloaded swh_storage fixture
# below

@pytest.fixture
def swh_storage():
    storage_config = {
        'cls': 'memory',
        'args': {
            'journal_writer': {
                'cls': 'memory',
            },
        },
    }
    storage = get_storage(**storage_config)
    return storage



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from typing import Dict, Generator, Sequence, Set

from swh.storage import get_storage


class FilteringProxyStorage:
    """Filtering Storage implementation. This is in charge of transparently
       filtering out known objects prior to adding them to storage.

    Sample configuration use case for filtering storage:

    .. code-block: yaml

        storage:
          cls: filter
          args:
            storage:
              cls: remote
              args: http://storage.internal.staging.swh.network:5002/

    """
    def __init__(self, storage):
        self.storage = get_storage(**storage)
        self.objects_seen = {
            'content': set(),    # set of content hashes (sha256) seen
            'directory': set(),
            'revision': set(),
        }

    def __getattr__(self, key):
        return getattr(self.storage, key)

    def content_add(self, content: Sequence[Dict]) -> Dict:
        contents = list(content)
        contents_to_add = self._filter_missing_contents(contents)
        return self.storage.content_add(
            x for x in contents if x['sha256'] in contents_to_add
        )

    def directory_add(self, directories: Sequence[Dict]) -> Dict:
        directories = list(directories)
        missing_ids = self._filter_missing_ids(
            'directory',
            (d['id'] for d in directories)
        )
        return self.storage.directory_add(
            d for d in directories if d['id'] in missing_ids
        )

    def revision_add(self, revisions):
        revisions = list(revisions)
        missing_ids = self._filter_missing_ids(
            'revision',
            (d['id'] for d in revisions)
        )
        return self.storage.revision_add(
            r for r in revisions if r['id'] in missing_ids
        )

    def _filter_missing_contents(
            self, content_hashes: Sequence[Dict]) -> Set[bytes]:
        """Return only the content keys missing from swh

        Args:
            content_hashes: List of sha256 to check for existence in swh
                storage

        """
        objects_seen = self.objects_seen['content']
        missing_hashes = []
        for hashes in content_hashes:
            if hashes['sha256'] in objects_seen:
                continue
            objects_seen.add(hashes['sha256'])
            missing_hashes.append(hashes)

        return set(self.storage.content_missing(
            missing_hashes,
            key_hash='sha256',
        ))

    def _filter_missing_ids(
            self,
            object_type: str,
            ids: Generator[bytes, None, None]) -> Set[bytes]:
        """Filter missing ids from the storage for a given object type.

        Args:
            object_type: object type to use {revision, directory}
            ids: Sequence of object_type ids

        Returns:
            Missing ids from the storage for object_type

        """
        objects_seen = self.objects_seen[object_type]
        missing_ids = []
        for id in ids:
            if id in objects_seen:
                continue
            objects_seen.add(id)
            missing_ids.append(id)

        fn_by_object_type = {
            'revision': self.storage.revision_missing,
            'directory': self.storage.directory_missing,
        }

        fn = fn_by_object_type[object_type]
        return set(fn(missing_ids))



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from unittest.mock import patch

from swh.storage import get_storage

from swh.storage.api.client import RemoteStorage
from swh.storage.storage import Storage as DbStorage
from swh.storage.in_memory import Storage as MemoryStorage
from swh.storage.buffer import BufferingProxyStorage
from swh.storage.filter import FilteringProxyStorage


@patch('swh.storage.storage.psycopg2.pool')
def test_get_storage(mock_pool):
    """Instantiating an existing storage should be ok

    """
    mock_pool.ThreadedConnectionPool.return_value = None
    for cls, real_class, dummy_args in [
            ('remote', RemoteStorage, {'url': 'url'}),
            ('memory', MemoryStorage, {}),
            ('local', DbStorage, {
                'db': 'postgresql://db', 'objstorage': {
                    'cls': 'memory', 'args': {},
                },
            }),
            ('filter', FilteringProxyStorage, {'storage': {
                'cls': 'memory', 'args': {}}
            }),
            ('buffer', BufferingProxyStorage, {'storage': {
                'cls': 'memory', 'args': {}}
            }),
    ]:
        actual_storage = get_storage(cls, args=dummy_args)
        assert actual_storage is not None
        assert isinstance(actual_storage, real_class)


def test_get_storage_failure():
    """Instantiating an unknown storage should raise

    """
    with pytest.raises(ValueError, match='Unknown storage class `unknown`'):
        get_storage('unknown', args=[])



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.storage.buffer import BufferingProxyStorage


def test_buffering_proxy_storage_content_threshold_not_hit(sample_data):
    contents = sample_data['content']
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'content': 10,
        }
    )
    s = storage.content_add([contents[0], contents[1]])
    assert s == {}

    # contents have not been written to storage
    missing_contents = storage.content_missing(
        [contents[0], contents[1]])
    assert set(missing_contents) == set(
        [contents[0]['sha1'], contents[1]['sha1']])

    s = storage.flush()
    assert s == {
        'content:add': 1 + 1,
        'content:add:bytes': contents[0]['length'] + contents[1]['length'],
        'skipped_content:add': 0
    }

    missing_contents = storage.content_missing(
        [contents[0], contents[1]])
    assert list(missing_contents) == []


def test_buffering_proxy_storage_content_threshold_nb_hit(sample_data):
    contents = sample_data['content']
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'content': 1,
        }
    )

    s = storage.content_add([contents[0]])
    assert s == {
        'content:add': 1,
        'content:add:bytes': contents[0]['length'],
        'skipped_content:add': 0
    }

    missing_contents = storage.content_missing([contents[0]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_content_threshold_bytes_hit(sample_data):
    contents = sample_data['content']
    content_bytes_min_batch_size = 20
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'content': 10,
            'content_bytes': content_bytes_min_batch_size,
        }
    )

    assert contents[0]['length'] > content_bytes_min_batch_size

    s = storage.content_add([contents[0]])
    assert s == {
        'content:add': 1,
        'content:add:bytes': contents[0]['length'],
        'skipped_content:add': 0
    }

    missing_contents = storage.content_missing([contents[0]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_directory_threshold_not_hit(sample_data):
    directories = sample_data['directory']
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'directory': 10,
        }
    )
    s = storage.directory_add([directories[0]])
    assert s == {}

    directory_id = directories[0]['id']
    missing_directories = storage.directory_missing(
        [directory_id])
    assert list(missing_directories) == [directory_id]

    s = storage.flush()
    assert s == {
        'directory:add': 1,
    }

    missing_directories = storage.directory_missing(
        [directory_id])
    assert list(missing_directories) == []


def test_buffering_proxy_storage_directory_threshold_hit(sample_data):
    directories = sample_data['directory']
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'directory': 1,
        }
    )
    s = storage.directory_add([directories[0]])
    assert s == {
        'directory:add': 1,
    }

    missing_directories = storage.directory_missing(
        [directories[0]['id']])
    assert list(missing_directories) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_revision_threshold_not_hit(sample_data):
    revisions = sample_data['revision']
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'revision': 10,
        }
    )
    s = storage.revision_add([revisions[0]])
    assert s == {}

    revision_id = revisions[0]['id']
    missing_revisions = storage.revision_missing(
        [revision_id])
    assert list(missing_revisions) == [revision_id]

    s = storage.flush()
    assert s == {
        'revision:add': 1,
    }

    missing_revisions = storage.revision_missing(
        [revision_id])
    assert list(missing_revisions) == []


def test_buffering_proxy_storage_revision_threshold_hit(sample_data):
    revisions = sample_data['revision']
    storage = BufferingProxyStorage(
        storage={'cls': 'memory', 'args': {}},
        min_batch_size={
            'revision': 1,
        }
    )
    s = storage.revision_add([revisions[0]])
    assert s == {
        'revision:add': 1,
    }

    missing_revisions = storage.revision_missing(
        [revisions[0]['id']])
    assert list(missing_revisions) == []

    s = storage.flush()
    assert s == {}



# Copyright (C) 2015-2016  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from . import storage

Storage = storage.Storage


class HashCollision(Exception):
    pass


STORAGE_IMPLEMENTATION = {'local', 'remote', 'memory', 'filter', 'buffer'}


def get_storage(cls, args):
    """Get a storage object of class `storage_class` with arguments
    `storage_args`.

    Args:
        storage (dict): dictionary with keys:
        - cls (str): storage's class, either local, remote, memory, filter,
            buffer
        - args (dict): dictionary with keys

    Returns:
        an instance of swh.storage.Storage (either local or remote)

    Raises:
        ValueError if passed an unknown storage class.

    """
    if cls not in STORAGE_IMPLEMENTATION:
        raise ValueError('Unknown storage class `%s`. Supported: %s' % (
            cls, ', '.join(STORAGE_IMPLEMENTATION)))

    if cls == 'remote':
        from .api.client import RemoteStorage as Storage
    elif cls == 'local':
        from .storage import Storage
    elif cls == 'memory':
        from .in_memory import Storage
    elif cls == 'filter':
        from .filter import FilteringProxyStorage as Storage
    elif cls == 'buffer':
        from .buffer import BufferingProxyStorage as Storage

    return Storage(**args)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Utility module to efficiently compute the list of changed files
# between two directory trees.
# The implementation is inspired from the work of Alberto Cortés
# for the go-git project. For more details, you can refer to:
#   - this blog post: https://blog.sourced.tech/post/difftree/
#   - the reference implementation in go:
#     https://github.com/src-d/go-git/tree/master/utils/merkletrie


import collections

from swh.model.hashutil import hash_to_bytes
from swh.model.identifiers import directory_identifier

from .dir_iterators import (
    DirectoryIterator, DoubleDirectoryIterator, Remaining
)

# get the hash identifier for an empty directory
_empty_dir_hash = hash_to_bytes(directory_identifier({'entries': []}))


def _get_rev(storage, rev_id):
    """
    Return revision data from swh storage.
    """
    return list(storage.revision_get([rev_id]))[0]


class _RevisionChangesList(object):
    """
    Helper class to track the changes between two
    revision directories.
    """

    def __init__(self, storage, track_renaming):
        """
        Args:
            storage: instance of swh storage
            track_renaming (bool): whether to track or not files renaming
        """
        self.storage = storage
        self.track_renaming = track_renaming
        self.result = []
        # dicts used to track file renaming based on hash value
        # we use a list instead of a single entry to handle the corner
        # case when a repository contains multiple instance of
        # the same file in different directories and a commit
        # renames all of them
        self.inserted_hash_idx = collections.defaultdict(list)
        self.deleted_hash_idx = collections.defaultdict(list)

    def add_insert(self, it_to):
        """
        Add a file insertion in the to directory.

        Args:
            it_to (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on the to directory
        """
        to_hash = it_to.current_hash()
        # if the current file hash has been previously marked as deleted,
        # the file has been renamed
        if self.track_renaming and self.deleted_hash_idx[to_hash]:
            # pop the delete change index in the same order it was inserted
            change = self.result[self.deleted_hash_idx[to_hash].pop(0)]
            # change the delete change as a rename one
            change['type'] = 'rename'
            change['to'] = it_to.current()
            change['to_path'] = it_to.current_path()
        else:
            # add the insert change in the list
            self.result.append({'type': 'insert',
                                'from': None,
                                'from_path': None,
                                'to': it_to.current(),
                                'to_path': it_to.current_path()})
            # if rename tracking is activated, add the change index in
            # the inserted_hash_idx dict
            if self.track_renaming:
                self.inserted_hash_idx[to_hash].append(len(self.result) - 1)

    def add_delete(self, it_from):
        """
        Add a file deletion in the from directory.

        Args:
            it_from (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on the from directory
        """
        from_hash = it_from.current_hash()
        # if the current file has been previously marked as inserted,
        # the file has been renamed
        if self.track_renaming and self.inserted_hash_idx[from_hash]:
            # pop the insert change index in the same order it was inserted
            change = self.result[self.inserted_hash_idx[from_hash].pop(0)]
            # change the insert change as a rename one
            change['type'] = 'rename'
            change['from'] = it_from.current()
            change['from_path'] = it_from.current_path()
        else:
            # add the delete change in the list
            self.result.append({'type': 'delete',
                                'from': it_from.current(),
                                'from_path': it_from.current_path(),
                                'to': None,
                                'to_path': None})
            # if rename tracking is activated, add the change index in
            # the deleted_hash_idx dict
            if self.track_renaming:
                self.deleted_hash_idx[from_hash].append(len(self.result) - 1)

    def add_modify(self, it_from, it_to):
        """
        Add a file modification in the to directory.

        Args:
            it_from (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on the from directory
            it_to (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on the to directory
        """
        self.result.append({'type': 'modify',
                            'from': it_from.current(),
                            'from_path': it_from.current_path(),
                            'to': it_to.current(),
                            'to_path': it_to.current_path()})

    def add_recursive(self, it, insert):
        """
        Recursively add changes from a directory.

        Args:
            it (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on a directory
            insert (bool): the type of changes to add (insertion
                or deletion)
        """
        # current iterated element is a regular file,
        # simply add adequate change in the list
        if not it.current_is_dir():
            if insert:
                self.add_insert(it)
            else:
                self.add_delete(it)
            return
        # current iterated element is a directory,
        dir_id = it.current_hash()
        # handle empty dir insertion/deletion as the swh model allow
        # to have such object compared to git
        if dir_id == _empty_dir_hash:
            if insert:
                self.add_insert(it)
            else:
                self.add_delete(it)
        # iterate on files reachable from it and add
        # adequate changes in the list
        else:
            sub_it = DirectoryIterator(self.storage, dir_id,
                                       it.current_path() + b'/')
            sub_it_current = sub_it.step()
            while sub_it_current:
                if not sub_it.current_is_dir():
                    if insert:
                        self.add_insert(sub_it)
                    else:
                        self.add_delete(sub_it)
                sub_it_current = sub_it.step()

    def add_recursive_insert(self, it_to):
        """
        Recursively add files insertion from a to directory.

        Args:
            it_to (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on a to directory
        """
        self.add_recursive(it_to, True)

    def add_recursive_delete(self, it_from):
        """
        Recursively add files deletion from a from directory.

        Args:
            it_from (swh.storage.algos.dir_iterators.DirectoryIterator):
                iterator on a from directory
        """
        self.add_recursive(it_from, False)


def _diff_elts_same_name(changes, it):
    """"
    Compare two directory entries with the same name and add adequate
    changes if any.

    Args:
        changes (_RevisionChangesList): the list of changes between
            two revisions
        it (swh.storage.algos.dir_iterators.DoubleDirectoryIterator):
            the iterator traversing two revision directories at the same time
    """
    # compare the two current directory elements of the iterator
    status = it.compare()
    # elements have same hash and same permissions:
    # no changes to add and call next on the two iterators
    if status['same_hash'] and status['same_perms']:
        it.next_both()
    # elements are regular files and have been modified:
    # insert the modification change in the list and
    # call next on the two iterators
    elif status['both_are_files']:
        changes.add_modify(it.it_from, it.it_to)
        it.next_both()
    # one element is a regular file, the other a directory:
    # recursively add delete/insert changes and call next
    # on the two iterators
    elif status['file_and_dir']:
        changes.add_recursive_delete(it.it_from)
        changes.add_recursive_insert(it.it_to)
        it.next_both()
    # both elements are directories:
    elif status['both_are_dirs']:
        # from directory is empty:
        # recursively add insert changes in the to directory
        # and call next on the two iterators
        if status['from_is_empty_dir']:
            changes.add_recursive_insert(it.it_to)
            it.next_both()
        # to directory is empty:
        # recursively add delete changes in the from directory
        # and call next on the two iterators
        elif status['to_is_empty_dir']:
            changes.add_recursive_delete(it.it_from)
            it.next_both()
        # both directories are not empty:
        # call step on the two iterators to descend further in
        # the directory trees.
        elif not status['from_is_empty_dir'] and not status['to_is_empty_dir']:
            it.step_both()


def _compare_paths(path1, path2):
    """
    Compare paths in lexicographic depth-first order.
    For instance, it returns:
        - "a" < "b"
        - "b/c/d" < "b"
        - "c/foo.txt" < "c.txt"
    """
    path1_parts = path1.split(b'/')
    path2_parts = path2.split(b'/')
    i = 0
    while True:
        if len(path1_parts) == len(path2_parts) and i == len(path1_parts):
            return 0
        elif len(path2_parts) == i:
            return 1
        elif len(path1_parts) == i:
            return -1
        else:
            if path2_parts[i] > path1_parts[i]:
                return -1
            elif path2_parts[i] < path1_parts[i]:
                return 1
        i = i + 1


def _diff_elts(changes, it):
    """
    Compare two directory entries.

    Args:
        changes (_RevisionChangesList): the list of changes between
            two revisions
        it (swh.storage.algos.dir_iterators.DoubleDirectoryIterator):
            the iterator traversing two revision directories at the same time
    """
    # compare current to and from path in depth-first lexicographic order
    c = _compare_paths(it.it_from.current_path(), it.it_to.current_path())
    # current from path is lower than the current to path:
    # the from path has been deleted
    if c < 0:
        changes.add_recursive_delete(it.it_from)
        it.next_from()
    # current from path is greater than the current to path:
    # the to path has been inserted
    elif c > 0:
        changes.add_recursive_insert(it.it_to)
        it.next_to()
    # paths are the same and need more processing
    else:
        _diff_elts_same_name(changes, it)


def diff_directories(storage, from_dir, to_dir, track_renaming=False):
    """
    Compute the differential between two directories, i.e. the list of
    file changes (insertion / deletion / modification / renaming)
    between them.

    Args:
        storage (swh.storage.storage.Storage): instance of a swh
            storage (either local or remote, for optimal performance
            the use of a local storage is recommended)
        from_dir (bytes): the swh identifier of the directory to compare from
        to_dir (bytes): the swh identifier of the directory to compare to
        track_renaming (bool): whether or not to track files renaming

    Returns:
        list: A list of dict representing the changes between the two
        revisions. Each dict contains the following entries:

            - *type*: a string describing the type of change
              ('insert' / 'delete' / 'modify' / 'rename')

            - *from*: a dict containing the directory entry metadata in the
              from revision (None in case of an insertion)

            - *from_path*: bytes string corresponding to the absolute path
              of the from revision entry (None in case of an insertion)

            - *to*: a dict containing the directory entry metadata in the
              to revision (None in case of a deletion)

            - *to_path*: bytes string corresponding to the absolute path
              of the to revision entry (None in case of a deletion)

        The returned list is sorted in lexicographic depth-first order
        according to the value of the *to_path* field.

    """
    changes = _RevisionChangesList(storage, track_renaming)
    it = DoubleDirectoryIterator(storage, from_dir, to_dir)
    while True:
        r = it.remaining()
        if r == Remaining.NoMoreFiles:
            break
        elif r == Remaining.OnlyFromFilesRemain:
            changes.add_recursive_delete(it.it_from)
            it.next_from()
        elif r == Remaining.OnlyToFilesRemain:
            changes.add_recursive_insert(it.it_to)
            it.next_to()
        else:
            _diff_elts(changes, it)
    return changes.result


def diff_revisions(storage, from_rev, to_rev, track_renaming=False):
    """
    Compute the differential between two revisions,
    i.e. the list of file changes between the two associated directories.

    Args:
        storage (swh.storage.storage.Storage): instance of a swh
            storage (either local or remote, for optimal performance
            the use of a local storage is recommended)
        from_rev (bytes): the identifier of the revision to compare from
        to_rev (bytes): the identifier of the revision to compare to
        track_renaming (bool): whether or not to track files renaming

    Returns:
        list: A list of dict describing the introduced file changes
        (see :func:`swh.storage.algos.diff.diff_directories`).

    """
    from_dir = None
    if from_rev:
        from_dir = _get_rev(storage, from_rev)['directory']
    to_dir = _get_rev(storage, to_rev)['directory']
    return diff_directories(storage, from_dir, to_dir, track_renaming)


def diff_revision(storage, revision, track_renaming=False):
    """
    Computes the differential between a revision and its first parent.
    If the revision has no parents, the directory to compare from
    is considered as empty.
    In other words, it computes the file changes introduced in a
    specific revision.

    Args:
        storage (swh.storage.storage.Storage): instance of a swh
            storage (either local or remote, for optimal performance
            the use of a local storage is recommended)
        revision (bytes): the identifier of the revision from which to
            compute the introduced changes.
        track_renaming (bool): whether or not to track files renaming

    Returns:
        list: A list of dict describing the introduced file changes
        (see :func:`swh.storage.algos.diff.diff_directories`).
    """
    rev_data = _get_rev(storage, revision)
    parent = None
    if rev_data['parents']:
        parent = rev_data['parents'][0]
    return diff_revisions(storage, parent, revision, track_renaming)
